const path = require('path')
const fs = require('fs')
const spawnSync = require('child_process').spawnSync

const protobufjsName = 'protobufjs'
const defaultTags = ['db', 'dbpost', 'dbpre', 'rpc']

async function foundGlobalProtobufjsPath(): Promise<string> {
  let target = ''
  // get yarn node_modules
  let res = spawnSync('yarn', ['global', 'dir'], { encoding: 'utf-8' })
  if (!res.stderr) {
    target = res.stdout.replace(/\n$/g, '')
    target = path.resolve(target, 'node_modules/', protobufjsName)
    if (fs.existsSync(target)) {
      return target
    }
  }

  // get npm node_modules
  res = spawnSync('npm', ['root', '-g'], { encoding: 'utf-8' })
  if (!res.stderr) {
    target = res.stdout.replace(/\n$/g, '')
    target = path.resolve(target, protobufjsName)
    if (fs.existsSync(target)) {
      return target
    }
  }

  return ''
}

exports.main = async (tags: string[]) => {
  // 先查找本地工程node_modules目录，不存在则查找全局node_modules目录
  let targetModule = path.resolve(__dirname, '../../../', protobufjsName)
  if (!fs.existsSync(targetModule)) {
    // 查找全局的yarn,npm包地址
    targetModule = await foundGlobalProtobufjsPath()
    // protobufjs模块未找到
    if (!targetModule) {
      console.log('can not found %s, please sure install it?', protobufjsName)
      return 1
    }
  }

  const keepTags: string[] = [...new Set([...defaultTags, ...tags])]
  // 进入需要修改模块的目录
  // protobufjs/cli/lib/pbjs.js#203行, 配置对象添加alternateCommentMode参数解析
  let workPath = path.resolve(targetModule, 'cli')
  if (fs.existsSync(workPath)) {
    let fileName = 'pbjs.js'
    let targetFile = path.resolve(workPath, fileName)
    const dataStr = fs.readFileSync(targetFile).toString('utf-8')
    try {
      const lines = dataStr.split(/\r?\n|\r/g)
      const startReg = /var parseOptions = {/
      const endReg = /};?/
      let startIndex = -1
      let endIndex = -1
      for (let i = 0; i < lines.length; i++) {
        const line = lines[i]
        if (startReg.test(line)) {
          startIndex = i + 1
          for (let j = i + 1; j < lines.length; j++) {
            if (endReg.test(lines[j])) {
              endIndex = j
              break
            }
          }
          break
        }
      }
      if (!(startIndex === -1 || endIndex === -1)) {
        const source = lines.slice(startIndex, endIndex)
        const hasAlternateCommentMode = source.some((line: string) => line.includes('alternateCommentMode'))
        if (!hasAlternateCommentMode) {
          const spaceStr = ' '.repeat(8)
          let res = lines.slice(0, startIndex).join('\n') + '\n'
          res += spaceStr + '"alternateCommentMode": argv["alternate-comment-mode"] || false,\n'
          res += lines.slice(startIndex, lines.length - 1).join('\n') + '\n'
          fs.writeFileSync(targetFile, res)
          console.log('set alternateCommentMode is true')
        }
      }
    } catch (e) {}
  }

  // static.js#132行，替换多级包名处理 push((config.es6 ? "const" : "var") + " " + escapeName(ns.name) + " = {};");
  workPath = path.resolve(targetModule, 'cli/targets')
  if (fs.existsSync(workPath)) {
    let fileName = 'static.js'
    let targetFile = path.resolve(workPath, fileName)
    const dataStr = fs.readFileSync(targetFile).toString('utf-8')
    const reg = /push\(\(config\.es6\s?\?\s?["']const["'].*escapeName\(ns\.name\).*=\s?{};["']\)/g
    if (reg.test(dataStr)) {
      const newLine = 'push((config.es6 ? "const" : "var") + " " + escapeName(ns.name) + " = $root." + escapeName(ns.name) + " || {};");'
      fs.writeFileSync(targetFile, dataStr.replace(reg, newLine))
      console.log('replace static.js escapeName')
    }
  }

  // 修改protobufjs下jsDoc插件json配置，允许不识别的tag
  workPath = path.resolve(targetModule, 'cli/lib')
  if (fs.existsSync(workPath)) {
    let fileName = 'tsd-jsdoc.json'
    let targetFile = path.resolve(workPath, fileName)
    const tsdJsdoc = fs.readFileSync(targetFile).toString('utf-8')
    try {
      const tsdJsdocObj = JSON.parse(tsdJsdoc)
      if (tsdJsdocObj.tags && !tsdJsdocObj.tags.allowUnknownTags) {
        tsdJsdocObj.tags.allowUnknownTags = true
        fs.writeFileSync(targetFile, JSON.stringify(tsdJsdocObj, null, 2))
        console.log('set allowUnknownTags is true')
      }
    } catch (e) {}
  }

  // 修改protobufjs下jsDoc插件publish.js, 保留需要的tag注释
  workPath = path.resolve(targetModule, 'cli/lib/tsd-jsdoc')
  if (fs.existsSync(workPath)) {
    let fileName = 'publish.js'
    let targetFile = path.resolve(workPath, fileName)
    const dataStr = fs.readFileSync(targetFile).toString('utf-8')
    try {
      const lines = dataStr.split(/\r?\n|\r/g)
      const startReg = /var keepTags = \[/
      const endReg = /\];?/
      let startIndex = -1
      let endIndex = -1
      for (let i = 0; i < lines.length; i++) {
        const line = lines[i]
        if (startReg.test(line)) {
          startIndex = i + 1
          for (let j = i + 1; j < lines.length; j++) {
            if (endReg.test(lines[j])) {
              endIndex = j
              break
            }
          }
          break
        }
      }
      if (!(startIndex === -1 || endIndex === -1)) {
        const spaceStr = ' '.repeat(4)
        let res = lines.slice(0, startIndex).join('\n') + '\n'
        const source = lines.slice(startIndex, endIndex)
        const addTags: string[] = []
        keepTags.forEach((tag) => {
          const hasTag = source.some((line: string) => line.includes(tag))
          if (!hasTag) {
            res += spaceStr + `"${tag}",\n`
            addTags.push(tag)
          }
        })
        res += lines.slice(startIndex, lines.length - 1).join('\n') + '\n'
        fs.writeFileSync(targetFile, res)
        addTags.length && console.log('add custom jsdoc tag', addTags)
      }
    } catch (e) {}
  }

}
